-- This file will contain functions needed for automatic self-maintenance when upgrading from previous versions.

local L = LibStub("AceLocale-3.0"):GetLocale("TrackOMatic", true);

--========================================
-- This function compares the ids of an
-- old version and a new version, and
-- returns whether the upgrade has
-- crossed a specific version id
--========================================
local function compareVersion(check, oldver, newver)
    if ((oldver < check) and (newver >= check)) then
        return true;
    end
    return false;
end

-- this swaps the old "show category headers" profile settings into "HIDE category headers"
function TrackOMatic_Upgrade_10005(oldver)
    if (compareVersion(10005, oldver, TRACKOMATIC_VERSIONID)) then
        TrackOMatic_Message(string.format(L["AUTO_DB_UPDATE"], 10005));
        for profile, data in pairs(TRACKOMATIC_VARS) do
            if (profile ~= "config") then
                local showHeaders = (data['show_category_headers'] or true);
                TRACKOMATIC_VARS[profile]['hide_category_headers'] = (not showHeaders);
                TRACKOMATIC_VARS[profile]['show_category_headers'] = nil;
            end
        end
    end
end

-- this automatically clears currency bars from previous versions that used string IDs instead of integer IDs
function TrackOMatic_Upgrade_10404(oldver)
    if (compareVersion(10404, oldver, TRACKOMATIC_VERSIONID)) then
        TrackOMatic_Message(string.format(L["AUTO_DB_UPDATE"], 10404));
        for profile, data in pairs(TRACKOMATIC_VARS) do
            if (profile ~= "config") then
                local showHeaders = (data['show_category_headers'] or true);
                TRACKOMATIC_VARS[profile]['categories']['currency'].entries = {};
            end
        end
    end
end
