local L = LibStub("AceLocale-3.0"):GetLocale("TrackOMatic", true);

TRACKOMATIC_CURRENCY_COLORS = {
    [392] = {
        ["Alliance"] = {r = 0.3, g = 0.3, b = 1},
        ["Horde"] = {r = 1, g = 0, b = 0},
    };
    [390] = {
        ["Alliance"] = {r = 0.3, g = 0.3, b = 1},
        ["Horde"] = {r = 1, g = 0, b = 0},
    };
    [395] = {r = 0, g = 0.5, b = 1},
    [515] = {r = 0.9, g = 0.25, b = 0.2},
    [396] = {r = 0.7, g = 0.6, b = 0.8},
    [61] = {r = 1, g = 0.25, b = 0.5},
    [391] = {r = 0.4, g = 0.7, b = 1},
    [402] = {r = 1, g = 0.25, b = 0},
    [241] = {r = 0.9, g = 0.7, b = 0},
    [416] = {r = 0.6, g = 0.4, b = 0.2},
    [738] = {r = 0.6, g = 0.6, b = 0.75},
    [752] = {r = 0.7, g = 0.6, b = 0.45},
    [776] = {r = 0.8, g = 0.3, b = 0.3},
    ["default"] = {r = 0.6, g = 0.5, b = 0}
};

local questIconNormal = "|TInterface\\GossipFrame\\AvailableQuestIcon:0:0:0:0|t";
local questIconDaily = "|TInterface\\GossipFrame\\DailyQuestIcon:0:0:0:0|t";

--========================================
-- Add currency with the specified
-- goal amount
--========================================
function TrackOMatic_AddCurrency(id, goalCount)

    local name = GetCurrencyInfo(id);

    -- makes sure the specified currency id is valid
    if (not name) then
        TrackOMatic_Message(string.format(L["INVALID_CURRENCY_ID"], id));
        return;
    end

    goalCount = TrackOMatic_StrToInt(goalCount);
    -- makes sure the goal value entered is a number and that number is greater than 0
    if (goalCount < 1) then
        StaticPopup_Show("TRACKOMATIC_INVALID_GOAL");
        return;
    end

    if (TrackOMatic_AddToCategory("currency", {id = id, goal = goalCount})) then
        TrackOMatic_Message(string.format(L["TRACKER_ADD_CURRENCY"], name));
    end

end

--========================================
-- Update a currency bar
--========================================
function TrackOMatic_SetCurrencyBar(data)

    local label, max, value, color;
    local levelLabel = "";
    local tooltipTitle = "";
    local tooltipText = "";
    local menu = {};
    local menuTitle = "";
    local isGlowing = false;

    local name, currentAmount, texture, earnedThisWeek, weeklyCap, totalCap, isDiscovered = GetCurrencyInfo(data.id);

    local playerLevel = UnitLevel("player");

    if (name) then
        -- set some values
        label = name;
        max = data.goal;
        value = currentAmount;
        local playerFaction = UnitFactionGroup("player");
        if (TRACKOMATIC_CURRENCY_COLORS[data.id]) then
            if (TRACKOMATIC_CURRENCY_COLORS[data.id][playerFaction]) then
                color = TRACKOMATIC_CURRENCY_COLORS[data.id][playerFaction];
            else
                color = TRACKOMATIC_CURRENCY_COLORS[data.id];
            end
        else
            color = TRACKOMATIC_CURRENCY_COLORS["default"];
        end

        tooltipTitle = name;
        tooltipText = string.format(L["CURRENT_AMOUNT"], currentAmount) .. "\n";
        tooltipText = tooltipText .. string.format(L["GOAL_AMOUNT"], data.goal) .. "\n";

        -- "x until goal" line
        if (currentAmount < max) then
            tooltipText = tooltipText .. "|cffff8000" .. string.format(L["UNTIL_GOAL"], (max - currentAmount)) .. "|r";
        else
            if (currentAmount > max) then
                tooltipText = tooltipText .. "|cff00ff00" .. string.format(L["PAST_GOAL"], (currentAmount - max)) .. "|r";
            else
                tooltipText = tooltipText .. "|cff00ff00" .. L["GOAL_REACHED"] .. "|r";
            end
        end

        local generalInfo = {};
        local dungeonFinderInfo = {};
        local conversionInfo = {};

        local currencyDetails = "";
        -- conversion calculator for justice points
        if (data.id == 395) then
            if (currentAmount < max) then
                table.insert(generalInfo, TrackOMatic_ConvertFromCurrencyString(L["CURRENCY_HONOR_POINTS"], (math.ceil((max - value) / 250) * 375)));
            end

            -- wotlk weekly raid quest
            if (playerLevel >= 80) then
                table.insert(generalInfo, questIconDaily .. L["WEEKLY_RAID_QUEST"] .. ": " .. TrackOMatic_CalculateJusticeGain(value, max, 138));
            end
            if (playerLevel == 90) then
                table.insert(generalInfo, L["MOP_HEROIC_DUNGEON_BOSSES"] .. ": " .. TrackOMatic_CalculateJusticeGain(value, max, 100));
            end
            -- dungeon finder
            if ((playerLevel >= 59) and (playerLevel <= 70)) then
                table.insert(dungeonFinderInfo, L["RANDOM_BC_DUNGEONS"] .. ": " .. TrackOMatic_CalculateJusticeGain(value, max, 5));
            end
            if ((playerLevel >= 69) and (playerLevel <= 80)) then
                table.insert(dungeonFinderInfo, L["RANDOM_WOTLK_DUNGEONS"] .. ": " .. TrackOMatic_CalculateJusticeGain(value, max, 12));
            end
            if ((playerLevel >= 80) and (playerLevel <= 85)) then
                table.insert(dungeonFinderInfo, L["RANDOM_CATA_DUNGEONS"] .. ": " .. TrackOMatic_CalculateJusticeGain(value, max, 70));
            end
            if ((playerLevel >= 85) and (playerLevel < 90)) then
                table.insert(dungeonFinderInfo, L["RANDOM_MOP_DUNGEONS"] .. ": " .. TrackOMatic_CalculateJusticeGain(value, max, 60));
            end
            -- conversion & upgrade
            if (value >= 500) then
                table.insert(conversionInfo, string.format(L["CONVERT_TO_CURRENCY"], "|cffffffff" .. L["CURRENCY_HONOR_POINTS"] .. "|cff20a0ff") .. " x" .. (math.floor(value / 500) * 250));
                if (value >= 1500) then
                    table.insert(conversionInfo, string.format(L["UPGRADE_GEAR"], "|cffffffff" .. math.floor(value / 750) .. "|cff20a0ff"));
                end
            end
        -- conversion calculator for honor points
        elseif (data.id == 392) then
            if (currentAmount < max) then
                table.insert(generalInfo, TrackOMatic_ConvertFromCurrencyString(L["CURRENCY_JUSTICE_POINTS"], (math.ceil((max - value) / 250) * 500)));
            end

            if (value >= 375) then
                table.insert(conversionInfo, string.format(L["CONVERT_TO_CURRENCY"], "|cffffffff" .. L["CURRENCY_JUSTICE_POINTS"] .. "|cff20a0ff") .. " x" .. (math.floor(value / 375) * 250));
                if (value >= 1500) then
                    table.insert(conversionInfo, string.format(L["UPGRADE_GEAR"], "|cffffffff" .. math.floor(value / 750) .. "|cff20a0ff"));
                end
            end
        -- conversion calculator and task counters for valor points
        elseif (data.id == 396) then
            table.insert(generalInfo, questIconDaily .. L["DAILY_QUESTS"] .. ": " .. math.max(0, math.ceil((max - value) / 5)));
            table.insert(dungeonFinderInfo, L["RANDOM_HEROICS"] .. ": " .. math.max(0, math.ceil((max - value) / 40)));
            table.insert(dungeonFinderInfo, L["RANDOM_SCENARIOS"] .. ": " .. math.max(0, math.ceil((max - value) / 20)));
            if (value >= 250) then
                table.insert(conversionInfo, string.format(L["CONVERT_TO_CURRENCY"], "|cffffffff" .. L["CURRENCY_CONQUEST_POINTS"] .. "|cff20a0ff") .. " x" .. (math.floor(value / 250) * 250));
                table.insert(conversionInfo, string.format(L["UPGRADE_GEAR"], "|cffffffff" .. math.floor(value / 250) .. "|cff20a0ff"));
            end
        -- upgrade counter for conquest points
        elseif (data.id == 390) then
            if (value >= 250) then
                table.insert(conversionInfo, string.format(L["UPGRADE_GEAR"], "|cffffffff" .. math.floor(value / 750)));
            end
        -- farming calculator for darkmoon tickets
        elseif (data.id == 515) then
            if (value < max) then
                table.insert(generalInfo, questIconNormal .. L["DARKMOON_PROFESSION_QUEST_SETS"] .. ": " .. math.ceil((max - value) / 20));
                table.insert(generalInfo, questIconNormal .. L["DARKMOON_ARTIFACT_SETS"] .. ": " .. math.ceil((max - value) / 80));
                table.insert(generalInfo, questIconDaily .. L["DARKMOON_GAMES_COLLECTIVE"] .. ": " .. math.ceil((max - value) / 5));
                table.insert(generalInfo, L["DARKMOON_MONTHS"] .. ": " .. math.ceil((max - value) / 135));
            end
        end

        if (#generalInfo > 0) then
            for i, d in pairs(generalInfo) do
                currencyDetails = currencyDetails .. d .. "\n";
            end
        end
        if (#dungeonFinderInfo > 0) then
            if (currencyDetails ~= "") then currencyDetails = currencyDetails .. "\n" .. L["DUNGEON_FINDER"] .. "\n"; end
            for i, d in pairs(dungeonFinderInfo) do
                currencyDetails = currencyDetails .. "   " .. d .. "\n";
            end
        end
        if (#conversionInfo > 0) then
            if (currencyDetails ~= "") then currencyDetails = currencyDetails .. "\n"; end
            for i, d in pairs(conversionInfo) do
                currencyDetails = currencyDetails .. "|cff20a0ff" .. d .. "\n";
            end
        end
        if (string.sub(currencyDetails, -1) == "\n") then currencyDetails = string.sub(currencyDetails, 1, string.len(currencyDetails) - 1); end

        if (currencyDetails ~= "") then
            tooltipText = tooltipText .. "\n\n" .. currencyDetails;
        end

        if (TRACKOMATIC_VARS[TRACKOMATIC_PROFILE]['show_quantity_currency']) then
            levelLabel = value .. "/" .. max .. " ";
        end
        if (TRACKOMATIC_VARS[TRACKOMATIC_PROFILE]['show_percent_currency']) then
            levelLabel = levelLabel .. "(" .. math.floor((value / max) * 100) .. "%)";
        end

        menuTitle = name;
        menu = {
            TrackOMatic_CommonMenuItem("CHANGE_GOAL", function(value) TrackOMatic_ChangeCurrencyGoal(data.id, value); end, name),
        };

        if (TRACKOMATIC_VARS['config']['currency_default_glow'] and (value >= max)) then
            isGlowing = true;
        end

    else
        max = 1;
        value = 1;
        color = {r = 0.5, g = 0.5, b = 0.5};
        label = L["UNDEFINED"];
    end

    return value, max, color, label, levelLabel, tooltipTitle, tooltipText, menuTitle, menu, nil, isGlowing;
end

--========================================
-- Removes a currency from the tracker
--========================================
function TrackOMatic_RemoveCurrency(index, isSilent)
    local data = TrackOMatic_RemoveFromCategory("currency", index);
    local name = GetCurrencyInfo(data.id);
    if ((not isSilent) and data) then
        TrackOMatic_Message(string.format(L["TRACKER_REMOVE_CURRENCY"], (name or L["UNDEFINED"])));
    end
end

--========================================
-- Changes a currency goal
--========================================
function TrackOMatic_ChangeCurrencyGoal(id, value)
    for index, entry in pairs(TRACKOMATIC_VARS[TRACKOMATIC_PROFILE]['categories']['currency'].entries) do
        if (entry.id == id) then
            value = TrackOMatic_StrToInt(value);
            if (value > 0) then
                TRACKOMATIC_VARS[TRACKOMATIC_PROFILE]['categories']['currency'].entries[index].goal = value;
                TrackOMatic_Message(string.format(L["CURRENCY_GOAL_CHANGED"], GetCurrencyInfo(id), value));
            else
                StaticPopup_Show("TRACKOMATIC_INVALID_GOAL");
                return;
            end
        end
    end
    TrackOMatic_UpdateTracker();
end


--========================================
-- Handler for clicking the add button
-- on the currency tab
--========================================
function TrackOMatic_AddCurrencyButton_OnClick(self)
    local name = TokenFrame.selectedToken;
    local id = TrackOMatic_GetCurrencyID(name);
    if (id) then
        TrackOMatic_ShowGoalInput(name, function(value) TrackOMatic_AddCurrency(id, value); end);
    else
        TrackOMatic_Message("|cffff0000" .. string.format(L["CURRENCY_ERROR"], name));
    end
end

--========================================
-- Retrieve the ID of the specified
-- currency
--========================================
function TrackOMatic_GetCurrencyID(name)

    for i = 1, 1000, 1 do
        local crName, currentAmount, texture, earnedThisWeek, weeklyCap, totalCap, isDiscovered = GetCurrencyInfo(i);
        if (crName == name) then
            return i;
        end
    end

    return nil;

end

function TrackOMatic_ConvertFromCurrencyString(currencyName, amount)
    local cvFrom = L["CONVERT_FROM_CURRENCY"];
    cvFrom = string.gsub(cvFrom, "{$a}", "|cffffffff" .. amount .. "|r");
    cvFrom = string.gsub(cvFrom, "{$c}", "|cffffffff" .. currencyName .. "|r");
    return cvFrom;
end

function TrackOMatic_CalculateJusticeGain(current, maxValue, baseGain)
    local multiplier = 1;
    return math.max(0, math.ceil((maxValue - current) / math.floor(baseGain * multiplier)));
end
